/*$Id: int64.c,v 1.11 2002/06/14 13:54:04 aa5779 Exp $*/
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <modules.h>
#include <interp.h>
#include <util.h>
#include <lists.h>
#include <dstring.h>
#include <limits.h>
#include <stdlib.h>
#include "version.h"
/// \section[NAME]
// \em[int64] -- a module for NSL (see \ref[sl(1)])

/// \section[OVERVIEW]
// This module extends standard NSL arithmetic builtins so that
// they operate on 64-bit values (strictly speaking, it depends
// on the size of long long on your system)
// 
// The functions affected are: isint, add, sub, mul, div, rem,
// incr, postincr, decr, postdecr, neg, abs, int, hex, dec, oct,
// all comparison functions, bit* functions.
// 
// Other functions (e.g. \em[expr]) remain 32-bit.
//
// The only builtin added is \em[uint32(value)] that behaves
// like \em[uint8] and \em[uint16]

#if SIZEOF_LONG_LONG != 0
static module *this_module;
extern procstack *get_pstack(void);
extern variable *safe_get_var(interpreter *i, char *name, int writing);

#define INT_LENGTH 0

static long long convert_to_ll(char *str)
{
	char *check = "";
	long long nv = *str ? strtoll(str + (*str == '#'), &check, *str == '#' ? 16 : 0) : 0LL;
	if(*check)
		recover_error("can't convert value '%s'", str);
	return nv;
}

void int64_LTX_init(module *cur_mod)
{
	this_module = cur_mod;
}

static void bi_isint(chunk *current, interpreter *i, value *args)
{
	char *str = DS_BODY(args->val);
	char *check = "";
	*str ? strtoll(str + (*str == '#'), &check, *str == '#' ? 16 : 0) : 0LL;
	if(!*check)
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_add(chunk *current, interpreter *i, value *args)
{
	long long value = 0;

	for(; args; args = list_next(args))
		value += convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(value, 10, INT_LENGTH));
}


static void bi_postincr(chunk *current, interpreter *i, value *args)
{
	long long val = 0;
   	long long ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_ll(DS_BODY(next_value(args)->val));
	ovalue = convert_to_ll(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromllint(val + ovalue, 10, 0));
}

static void bi_incr(chunk *current, interpreter *i, value *args)
{
	long long val = 0; 
	long long ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	val = convert_to_ll(DS_BODY(next_value(args)->val));
	ovalue = convert_to_ll(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromllint(val + ovalue, 10, 0));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

static void bi_postdecr(chunk *current, interpreter *i, value *args)
{
	long long val = 0;
   	long long ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_ll(DS_BODY(next_value(args)->val));
	ovalue = convert_to_ll(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromllint(ovalue - val, 10, 0));
}

static void bi_decr(chunk *current, interpreter *i, value *args)
{
	long long val = 0;
	long long ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ovalue = convert_to_ll(DS_BODY(var->vals->val));
	val = convert_to_ll(DS_BODY(next_value(args)->val));
	ds_assign(var->vals->val, ds_fromllint(ovalue - val, 10, 0));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

static void bi_mul(chunk *current, interpreter *i, value *args)
{
	long long value = 1;

	for(; args; args = list_next(args))
		value *= convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(value, 10, INT_LENGTH));
}

static void bi_sub(chunk *current, interpreter *i, value *args)
{
	long long v1 = convert_to_ll(DS_BODY(args->val));
	long long v2 = convert_to_ll(DS_BODY(((value	*)list_next(args))->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(v1 - v2, 10, INT_LENGTH));
}

static void bi_div(chunk *current, interpreter *i, value *args)
{
	long long v1 = convert_to_ll(DS_BODY(args->val));
	long long v2 = convert_to_ll(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(get_pstack()->state->buffer, ds_fromllint(v1 / v2, 10, INT_LENGTH));
}

static void bi_rem(chunk *current, interpreter *i, value *args)
{
	long long v1 = convert_to_ll(DS_BODY(args->val));
	long long v2 = convert_to_ll(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(get_pstack()->state->buffer, ds_fromllint(v1 % v2, 10, INT_LENGTH));
}

static void bi_neg(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(-v, 10, INT_LENGTH));
}

static void bi_abs(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(llabs(v), 10, INT_LENGTH));
}

static void bi_int(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	int shift;
	static char hd[] = "0123456789ABCDEF";
	ds_appendch(get_pstack()->state->buffer, '#');
	if(v >= 0)
		ds_appendch(get_pstack()->state->buffer, '0');
	else
	{
		ds_appendch(get_pstack()->state->buffer, '-');
		v = -v;
	}
	for(shift = 60; shift >=0; shift -= 4) // mind unportable magic numbers
		ds_appendch(get_pstack()->state->buffer, hd[(v >> shift) & 0x0F]);
}

static void bi_dec(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(v, 10, INT_LENGTH));
}

static void bi_hex(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(v, 16, INT_LENGTH));
}

static void bi_oct(chunk *current, interpreter *i, value *args)
{
	long long v = convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(v, 8, INT_LENGTH));
}

static void pass_thru(chunk *current, interpreter *i, value *args, int opcode)
{
	instruction_t *old_vector = get_ma_data(this_module, i);
	if(old_vector)
		old_vector[opcode](current, i, args);
}

enum opcodes_in_question { op_eq = 16, op_neq, op_less, op_greater, op_le, op_ge, op_between };
// (refer to the following instruction table)

static void bi_eq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_eq);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val))  == convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_neq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_neq);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val)) != convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_less(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_less);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val)) < convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_greater(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_greater);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val)) > convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_le(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_le);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val)) <= convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_ge(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_ge);
	else
	if(convert_to_ll(DS_BODY(next_value(args)->val)) >= convert_to_ll(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_between(chunk *current, interpreter *i, value *args)
{
	if(!next_value(third_value(args)) || strcmp(DS_BODY(args->val), "int"))
		pass_thru(current, i, args, op_between);
	else
	{
		long long i1 = convert_to_ll(DS_BODY(next_value(args)->val));
		long long i2 = convert_to_ll(DS_BODY(third_value(args)->val));
		long long i3 = convert_to_ll(DS_BODY(next_value(third_value(args))->val));
		if(i1 <= i2 && i2 <= i3)
			ds_appendstr(get_pstack()->state->buffer, "true");
	}
}

static void bi_bitand(chunk *current, interpreter *i, value *args)
{
	long long val;

	val = convert_to_ll(DS_BODY(args->val));
	for(args = list_next(args); args; args = list_next(args))
		val &= convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(val, 10, INT_LENGTH));
}

static void bi_bitor(chunk *current, interpreter *i, value *args)
{
	long long val = 0;

	for(; args; args = list_next(args))
		val |= convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(val, 10, INT_LENGTH));
}

static void bi_bitxor(chunk *current, interpreter *i, value *args)
{
	long long val = 0;

	for(; args; args = list_next(args))
		val ^= convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(val, 10, INT_LENGTH));
}

static void bi_bitinv(chunk *current, interpreter *i, value *args)
{
	long long val = ~convert_to_ll(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(val, 10, INT_LENGTH));
}

static void bi_bitshift(chunk *current, interpreter *i, value *args)
{
	long long val = convert_to_ll(DS_BODY(args->val));
	int shift = convert_to_int(DS_BODY(next_value(args)->val));

	if(shift < 0)
		val >>= -shift;
	else 
		val <<= shift;
	ds_append(get_pstack()->state->buffer, ds_fromllint(val, 10, INT_LENGTH));
}

static void bi_uint32(chunk *current, interpreter *i, value *args)
{
	unsigned long v = (signed long)convert_to_int(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromllint(v, 10, INT_LENGTH));
}

static master_instruction_src int64_add_builtins[] = {
							{"uint32", 1, bi_uint32, 1, 1, 1, 1},
							{NULL}
							};

static master_instruction_src int64_builtins[] = {
							{"isint", 1, bi_isint, 1, 1, 1, 1},
							{"add", 1, bi_add, 2, INT_MAX, 2, 1},
							{"sub", 1, bi_sub, 2, 2, 2, 1},
							{"mul", 1, bi_mul, 2, INT_MAX, 2, 1},
							{"div", 1, bi_div, 2, 2, 2, 1},
							{"rem", 1, bi_rem, 2, 2, 2, 1},
							{"neg", 1, bi_neg, 1, 1, 1, 1},
							{"abs", 1, bi_abs, 1, 1, 1, 1},
							{"int", 1, bi_int, 1, 1, 1, 1},
							{"incr", 1, bi_incr, 2, 2, 2},
							{"decr", 1, bi_decr, 2, 2, 2},
							{"postincr", 1, bi_postincr, 2, 2, 2},
							{"postdecr", 1, bi_postdecr, 2, 2, 2},
							{"dec", 1, bi_dec, 1, 1, 1, 1},
							{"hex", 1, bi_hex, 1, 1, 1, 1},
							{"oct", 1, bi_oct, 1, 1, 1, 1},
							{"eq", 1, bi_eq, 2, 3, 2, 1},
							{"neq", 1, bi_neq, 2, 3, 2, 1},
							{"less", 1, bi_less, 2, 3, 2, 1},
							{"greater", 1, bi_greater, 2, 3, 2, 1},
							{"le", 1, bi_le, 2, 3, 2, 1},
							{"ge", 1, bi_ge, 2, 3, 2, 1},
							{"between", 1, bi_between, 3, 4, 3, 1},
							{"bitand", 1, bi_bitand, 2, INT_MAX, 2, 1},
							{"bitor", 1, bi_bitor, 2, INT_MAX, 2, 1},
							{"bitxor", 1, bi_bitxor, 2, INT_MAX, 2, 1},
							{"bitinv", 1, bi_bitinv, 1, 1, 1, 1},
							{"bitshift", 1, bi_bitshift, 2, 2, 2, 1},
							{NULL}
							};

void int64_LTX_attach(mod_attach *ma)
{
	 ma->data = xmalloc(sizeof(int64_builtins) * sizeof(instruction_t) / sizeof(*int64_builtins));
	 replace_builtins(ma->i, int64_builtins, ma->data);
	 append_builtins(ma->i, int64_add_builtins);
}

const char int64_LTX_version[] = VERSION;
const char int64_LTX_date[] = __DATE__;

#endif
/// \section[AUTHOR]
// Artem V. Andreev

/// \section[COPYRIGHT]
// see \ref[sl(1)]

/// \seealso
